@@ -5,6 +5,7 @@ from __future__ import annotations
 import asyncio
 import base64
 import binascii
+import gzip
 import hashlib
 import hmac
 import json
@@ -13,20 +14,19 @@ import math
 import secrets
 import struct
 import time
-from typing import Any
+from typing import Any, Callable
 
 import aiohttp
 from Crypto.Cipher import AES
 from Crypto.Util.Padding import pad, unpad
 
 from roborock.exceptions import (
-    RoborockException,
+    RoborockException, RoborockTimeout, VacuumError,
 )
 from .code_mappings import WASH_MODE_MAP, DUST_COLLECTION_MAP, RoborockDockType, \
-    RoborockDockDustCollectionType, RoborockDockWashingModeType
+    RoborockDockDustCollectionType, RoborockDockWashingModeType, STATE_CODE_TO_STATUS
 from .containers import (
     UserData,
-    HomeDataDevice,
     Status,
     CleanSummary,
     Consumable,
@@ -37,6 +37,7 @@ from .containers import (
     SmartWashParameters,
 
 )
+from .roborock_queue import RoborockQueue
 from .typing import (
     RoborockDeviceProp,
     RoborockCommand,
@@ -46,6 +47,9 @@ from .typing import (
 _LOGGER = logging.getLogger(__name__)
 QUEUE_TIMEOUT = 4
 MQTT_KEEPALIVE = 60
+SPECIAL_COMMANDS = [
+    RoborockCommand.GET_MAP_V1,
+]
 
 
 def md5hex(message: str) -> str:
@@ -86,7 +90,7 @@ class PreparedRequest:
                 return await resp.json()
 
 
-class RoborockClient():
+class RoborockClient:
 
     def __init__(self, endpoint: str, device_localkey: dict[str, str], prefixed=False) -> None:
         self.device_localkey = device_localkey
@@ -97,12 +101,14 @@ class RoborockClient():
         self._endpoint = base64.b64encode(md5bin(endpoint)[8:14]).decode()
         self._nonce = secrets.token_bytes(16)
         self._prefixed = prefixed
+        self._waiting_queue: dict[int, RoborockQueue] = {}
+        self._status_listeners: list[Callable[[str, str], None]] = []
 
     def _decode_msg(self, msg: bytes, local_key: str) -> dict[str, Any]:
         if self._prefixed:
             msg = msg[4:]
         if msg[0:3] != "1.0".encode():
-            raise RoborockException("Unknown protocol version")
+            raise RoborockException(f"Unknown protocol version {msg[0:3]}")
         if len(msg) == 17:
             [version, _seq, _random, timestamp, protocol] = struct.unpack(
                 "!3sIIIH", msg[0:17]
@@ -112,13 +118,19 @@ class RoborockClient():
                 "timestamp": timestamp,
                 "protocol": protocol,
             }
-        crc32 = binascii.crc32(msg[0: len(msg) - 4])
+        # crc32 = binascii.crc32(msg[0: len(msg) - 4])
         [version, _seq, _random, timestamp, protocol, payload_len] = struct.unpack(
             "!3sIIIHH", msg[0:19]
         )
+        if not payload_len:
+            return {
+                "version": version,
+                "timestamp": timestamp,
+                "protocol": protocol,
+            }
         [payload, expected_crc32] = struct.unpack_from(f"!{payload_len}sI", msg, 19)
-        if crc32 != expected_crc32:
-            raise RoborockException(f"Wrong CRC32 {crc32}, expected {expected_crc32}")
+        # if crc32 != expected_crc32:
+        #     raise RoborockException(f"Wrong CRC32 {crc32}, expected {expected_crc32}")
 
         aes_key = md5bin(encode_timestamp(timestamp) + local_key + self._salt)
         decipher = AES.new(aes_key, AES.MODE_ECB)
@@ -130,7 +142,7 @@ class RoborockClient():
             "payload": decrypted_payload,
         }
 
-    def _get_msg_raw(self, device_id, protocol, timestamp, payload, prefix='') -> bytes:
+    def _encode_msg(self, device_id, protocol, timestamp, payload, prefix='') -> bytes:
         local_key = self.device_localkey[device_id]
         aes_key = md5bin(encode_timestamp(timestamp) + local_key + self._salt)
         cipher = AES.new(aes_key, AES.MODE_ECB)
@@ -155,8 +167,85 @@ class RoborockClient():
         msg += struct.pack("!I", crc32)
         return msg
 
+    async def on_message(self, device_id, msg) -> bool:
+        try:
+            data = self._decode_msg(msg, self.device_localkey[device_id])
+            protocol = data.get("protocol")
+            if protocol == 102 or protocol == 4:
+                payload = json.loads(data.get("payload").decode())
+                for data_point_number, data_point in payload.get("dps").items():
+                    if data_point_number == "102":
+                        data_point_response = json.loads(data_point)
+                        request_id = data_point_response.get("id")
+                        queue = self._waiting_queue.get(request_id)
+                        if queue:
+                            if queue.protocol == protocol:
+                                error = data_point_response.get("error")
+                                if error:
+                                    await queue.async_put(
+                                        (
+                                            None,
+                                            VacuumError(
+                                                error.get("code"), error.get("message")
+                                            ),
+                                        ),
+                                        timeout=QUEUE_TIMEOUT,
+                                    )
+                                else:
+                                    result = data_point_response.get("result")
+                                    if isinstance(result, list) and len(result) == 1:
+                                        result = result[0]
+                                    await queue.async_put(
+                                        (result, None), timeout=QUEUE_TIMEOUT
+                                    )
+                                    return True
+                        elif request_id < self._id_counter:
+                            _LOGGER.debug(
+                                f"id={request_id} Ignoring response: {data_point_response}"
+                            )
+                    elif data_point_number == "121":
+                        status = STATE_CODE_TO_STATUS.get(data_point)
+                        _LOGGER.debug(f"Status updated to {status}")
+                        for listener in self._status_listeners:
+                            listener(device_id, status)
+                    else:
+                        _LOGGER.debug(
+                            f"Unknown data point number received {data_point_number} with {data_point}"
+                        )
+            elif protocol == 301:
+                payload = data.get("payload")[0:24]
+                [endpoint, _, request_id, _] = struct.unpack("<15sBH6s", payload)
+                if endpoint.decode().startswith(self._endpoint):
+                    iv = bytes(AES.block_size)
+                    decipher = AES.new(self._nonce, AES.MODE_CBC, iv)
+                    decrypted = unpad(
+                        decipher.decrypt(data.get("payload")[24:]), AES.block_size
+                    )
+                    decrypted = gzip.decompress(decrypted)
+                    queue = self._waiting_queue.get(request_id)
+                    if queue:
+                        if isinstance(decrypted, list):
+                            decrypted = decrypted[0]
+                        await queue.async_put((decrypted, None), timeout=QUEUE_TIMEOUT)
+                        return True
+        except Exception as ex:
+            _LOGGER.exception(ex)
+
+    async def _async_response(self, request_id: int, protocol_id: int = 0) -> tuple[Any, VacuumError | None]:
+        try:
+            queue = RoborockQueue(protocol_id)
+            self._waiting_queue[request_id] = queue
+            (response, err) = await queue.async_get(QUEUE_TIMEOUT)
+            return response, err
+        except (asyncio.TimeoutError, asyncio.CancelledError):
+            raise RoborockTimeout(
+                f"Timeout after {QUEUE_TIMEOUT} seconds waiting for response"
+            ) from None
+        finally:
+            del self._waiting_queue[request_id]
+
     def _get_payload(
-            self, method: RoborockCommand, params: list = None
+            self, method: RoborockCommand, params: list = None, secured=False
     ):
         timestamp = math.floor(time.time())
         request_id = self._id_counter
@@ -165,11 +254,12 @@ class RoborockClient():
             "id": request_id,
             "method": method,
             "params": params or [],
-            "security": {
-                "endpoint": self._endpoint,
-                "nonce": self._nonce.hex().upper(),
-            },
         }
+        if secured:
+            inner["security"] = {
+                "endpoint": self._endpoint,
+                "nonce": "39344139463753454b4851444f4a4442",
+            }
         payload = bytes(
             json.dumps(
                 {
@@ -192,53 +282,75 @@ class RoborockClient():
             return Status(status)
 
     async def get_dnd_timer(self, device_id: str) -> DNDTimer:
-        dnd_timer = await self.send_command(device_id, RoborockCommand.GET_DND_TIMER)
-        if isinstance(dnd_timer, dict):
-            return DNDTimer(dnd_timer)
+        try:
+            dnd_timer = await self.send_command(device_id, RoborockCommand.GET_DND_TIMER)
+            if isinstance(dnd_timer, dict):
+                return DNDTimer(dnd_timer)
+        except RoborockTimeout as e:
+            _LOGGER.error(e)
 
     async def get_clean_summary(self, device_id: str) -> CleanSummary:
-        clean_summary = await self.send_command(
-            device_id, RoborockCommand.GET_CLEAN_SUMMARY
-        )
-        if isinstance(clean_summary, dict):
-            return CleanSummary(clean_summary)
-        elif isinstance(clean_summary, bytes):
-            return CleanSummary({"clean_time": clean_summary})
+        try:
+            clean_summary = await self.send_command(
+                device_id, RoborockCommand.GET_CLEAN_SUMMARY
+            )
+            if isinstance(clean_summary, dict):
+                return CleanSummary(clean_summary)
+            elif isinstance(clean_summary, bytes):
+                return CleanSummary({"clean_time": clean_summary})
+        except RoborockTimeout as e:
+            _LOGGER.error(e)
 
     async def get_clean_record(self, device_id: str, record_id: int) -> CleanRecord:
-        clean_record = await self.send_command(
-            device_id, RoborockCommand.GET_CLEAN_RECORD, [record_id]
-        )
-        if isinstance(clean_record, dict):
-            return CleanRecord(clean_record)
+        try:
+            clean_record = await self.send_command(
+                device_id, RoborockCommand.GET_CLEAN_RECORD, [record_id]
+            )
+            if isinstance(clean_record, dict):
+                return CleanRecord(clean_record)
+        except RoborockTimeout as e:
+            _LOGGER.error(e)
 
     async def get_consumable(self, device_id: str) -> Consumable:
-        consumable = await self.send_command(device_id, RoborockCommand.GET_CONSUMABLE)
-        if isinstance(consumable, dict):
-            return Consumable(consumable)
+        try:
+            consumable = await self.send_command(device_id, RoborockCommand.GET_CONSUMABLE)
+            if isinstance(consumable, dict):
+                return Consumable(consumable)
+        except RoborockTimeout as e:
+            _LOGGER.error(e)
 
     async def get_washing_mode(self, device_id: str) -> RoborockDockWashingModeType:
-        washing_mode = await self.send_command(device_id, RoborockCommand.GET_WASH_TOWEL_MODE)
-        return WASH_MODE_MAP.get(washing_mode['wash_mode'])
+        try:
+            washing_mode = await self.send_command(device_id, RoborockCommand.GET_WASH_TOWEL_MODE)
+            return WASH_MODE_MAP.get(washing_mode['wash_mode'])
+        except RoborockTimeout as e:
+            _LOGGER.error(e)
 
     async def get_dust_collection_mode(self, device_id: str) -> RoborockDockDustCollectionType:
-        dust_collection = await self.send_command(device_id, RoborockCommand.GET_DUST_COLLECTION_MODE)
-        return DUST_COLLECTION_MAP.get(dust_collection['mode'])
+        try:
+            dust_collection = await self.send_command(device_id, RoborockCommand.GET_DUST_COLLECTION_MODE)
+            return DUST_COLLECTION_MAP.get(dust_collection['mode'])
+        except RoborockTimeout as e:
+            _LOGGER.error(e)
 
     async def get_mop_wash_mode(self, device_id: str) -> SmartWashParameters:
-        mop_wash_mode = await self.send_command(device_id, RoborockCommand.GET_SMART_WASH_PARAMS)
-        if isinstance(mop_wash_mode, dict):
-            return SmartWashParameters(mop_wash_mode)
+        try:
+            mop_wash_mode = await self.send_command(device_id, RoborockCommand.GET_SMART_WASH_PARAMS)
+            if isinstance(mop_wash_mode, dict):
+                return SmartWashParameters(mop_wash_mode)
+        except RoborockTimeout as e:
+            _LOGGER.error(e)
 
     async def get_dock_summary(self, device_id: str, dock_type: RoborockDockType) -> RoborockDockSummary:
-        collection_mode = await self.get_dust_collection_mode(device_id)
-        mop_wash = None
-        washing_mode = None
-        if dock_type == RoborockDockType.EMPTY_WASH_FILL_DOCK:
-            [mop_wash, washing_mode] = await asyncio.gather(
-                *[self.get_mop_wash_mode(device_id), self.get_washing_mode(device_id)])
+        try:
+            commands = [self.get_dust_collection_mode(device_id)]
+            if dock_type == RoborockDockType.EMPTY_WASH_FILL_DOCK:
+                commands += [self.get_mop_wash_mode(device_id), self.get_washing_mode(device_id)]
+            [collection_mode, mop_wash, washing_mode] = (list(await asyncio.gather(*commands)) + [None, None])[:3]
 
-        return RoborockDockSummary(collection_mode, washing_mode, mop_wash)
+            return RoborockDockSummary(collection_mode, washing_mode, mop_wash)
+        except RoborockTimeout as e:
+            _LOGGER.error(e)
 
     async def get_prop(self, device_id: str) -> RoborockDeviceProp:
         [status, dnd_timer, clean_summary, consumable] = await asyncio.gather(
@@ -249,17 +361,9 @@ class RoborockClient():
                 self.get_consumable(device_id),
             ]
         )
-        last_clean_record = None
-        if clean_summary and clean_summary.records and len(clean_summary.records) > 0:
-            last_clean_record = await self.get_clean_record(
-                device_id, clean_summary.records[0]
-            )
-        dock_summary = None
-        if status and status.dock_type != RoborockDockType.NO_DOCK:
-            dock_summary = await self.get_dock_summary(device_id, status.dock_type)
         if any([status, dnd_timer, clean_summary, consumable]):
             return RoborockDeviceProp(
-                status, dnd_timer, clean_summary, consumable, last_clean_record, dock_summary
+                status, dnd_timer, clean_summary, consumable
             )
 
     async def get_multi_maps_list(self, device_id) -> MultiMapsList:
